@@ -15,6 +15,8 @@
 #include "util.h"
 #include "version.h"
 #include "image_processing.h"
+#include "clipboard.h"
+#include "print.h"
 
 void
 on_mainwindow_realize                  (GtkWidget       *widget,
@@ -48,7 +50,10 @@ void
 on_open_menu_activate                  (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   BEGIN_BUSY_CURSOR
    open_canvas(widget_get_image(GTK_WIDGET(menuitem)));
+   END_BUSY_CURSOR
 }
 
 
@@ -56,8 +61,10 @@ void
 on_save_menu_activate                  (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
-   save_canvas(widget_get_image(GTK_WIDGET(menuitem)));
-
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   BEGIN_BUSY_CURSOR
+   save_canvas(widget_get_image(GTK_WIDGET(menuitem)), 0);
+   END_BUSY_CURSOR
 }
 
 
@@ -65,6 +72,10 @@ void
 on_save_as_menu_activate               (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   BEGIN_BUSY_CURSOR
+   save_canvas(widget_get_image(GTK_WIDGET(menuitem)), 1);
+   END_BUSY_CURSOR
 
 }
 
@@ -73,6 +84,11 @@ void
 on_print_menu_activate                 (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   
+   do_print(ibuf);
+   
+
 
 }
 
@@ -100,6 +116,10 @@ void
 on_cut_menu_activate                   (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   g_assert(ibuf);
+   image_buf_cut(ibuf);
+
 
 }
 
@@ -108,6 +128,9 @@ void
 on_copy_menu_activate                  (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   g_assert(ibuf);
+   image_buf_copy(ibuf);
 
 }
 
@@ -116,7 +139,9 @@ void
 on_paste_menu_activate                 (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
-
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   g_assert(ibuf);
+   image_buf_paste(ibuf);
 }
 
 
@@ -124,7 +149,12 @@ void
 on_clear_menu_activate                 (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
-
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   g_assert(ibuf);
+   if (ibuf->num_pts)
+      image_buf_delete(ibuf);
+   else
+      image_buf_clear(ibuf);
 }
 
 
@@ -444,6 +474,8 @@ void
 on_change_background_menu_activate     (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
 {
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   image_buf_set_desktop_background(ibuf, 0);
 
 }
 
@@ -560,7 +592,10 @@ void
 on_open_button_clicked                 (GtkButton       *button,
                                         gpointer         user_data)
 {
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(button));
+   BEGIN_BUSY_CURSOR
    open_canvas(widget_get_image(GTK_WIDGET(button)));
+   END_BUSY_CURSOR
 }
 
 
@@ -568,8 +603,10 @@ void
 on_save_button_clicked                 (GtkButton       *button,
                                         gpointer         user_data)
 {
-   save_canvas(widget_get_image(GTK_WIDGET(button)));
-
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(button));
+   BEGIN_BUSY_CURSOR
+   save_canvas(widget_get_image(GTK_WIDGET(button)), 0);
+   END_BUSY_CURSOR
 }
 
 
@@ -577,6 +614,10 @@ void
 on_save_as_button_clicked              (GtkButton       *button,
                                         gpointer         user_data)
 {
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(button));
+   BEGIN_BUSY_CURSOR
+   save_canvas(widget_get_image(GTK_WIDGET(button)), 1);
+   END_BUSY_CURSOR
 
 }
 
@@ -585,6 +626,10 @@ void
 on_print_button_clicked                (GtkButton       *button,
                                         gpointer         user_data)
 {
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(button));
+   
+   do_print(ibuf);
+  
 
 }
 
@@ -736,6 +781,27 @@ on_drawingarea_motion_notify_event     (GtkWidget       *widget,
 }
 
 
+gboolean
+on_drawingarea_focus_in_event          (GtkWidget       *widget,
+                                        GdkEventFocus   *event,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(widget);
+   image_buf_focus_obtained(ibuf);   
+   return FALSE;
+}
+
+
+gboolean
+on_drawingarea_focus_out_event         (GtkWidget       *widget,
+                                        GdkEventFocus   *event,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(widget);
+   image_buf_focus_lost(ibuf);   
+   return FALSE;
+}
+
 void
 on_drawingarea_realize                 (GtkWidget       *widget,
                                         gpointer         user_data)
@@ -803,7 +869,7 @@ on_new_canvas_ok_button_clicked        (GtkButton       *button,
       gtk_widget_destroy(window);
 
 
-      create_image_buf("untitled", width, height);
+      create_image_buf(0, width, height);
    }
    else
    {
@@ -811,6 +877,7 @@ on_new_canvas_ok_button_clicked        (GtkButton       *button,
       gtk_window_set_modal(GTK_WINDOW(msgbox), TRUE);
        /* And mark it as a transient dialog */
       gtk_window_set_transient_for (GTK_WINDOW(msgbox), GTK_WINDOW(window));
+      gtk_window_set_position(GTK_WINDOW(msgbox), GTK_WIN_POS_CENTER);
       gnome_dialog_run(GNOME_DIALOG(msgbox));
    }   
 }
@@ -820,7 +887,8 @@ void
 on_new_canvas_cancel_button_clicked    (GtkButton       *button,
                                         gpointer         user_data)
 {
-
+   GtkWidget *window = widget_get_toplevel_parent(GTK_WIDGET(button)); 
+   gtk_widget_destroy(window);
 }
 
 
@@ -855,7 +923,7 @@ on_fontpicker_font_set                 (GnomeFontPicker *gnomefontpicker,
 
    image_buf *ibuf = widget_get_image(GTK_WIDGET(gnomefontpicker));
    g_assert(ibuf);
-   image_buf_set_font(ibuf, gnome_font_picker_get_font(gnomefontpicker));
+   image_buf_set_font(ibuf, gnome_font_picker_get_font(gnomefontpicker), gnome_font_picker_get_font_name(gnomefontpicker));
    
 
 }
@@ -1309,3 +1377,355 @@ on_drawingarea_key_release_event     (GtkWidget       *widget,
    return FALSE;
 }
 
+
+gboolean
+on_mainwindow_focus_in_event           (GtkWidget       *widget,
+                                        GdkEventFocus   *event,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(widget));
+   gtk_widget_grab_focus(ibuf->drawing_area);
+   gtk_widget_grab_default(ibuf->drawing_area);
+
+  return FALSE;
+}
+
+
+gboolean
+on_mainwindow_focus_out_event          (GtkWidget       *widget,
+                                        GdkEventFocus   *event,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(widget));
+   image_buf_focus_lost(ibuf);
+  return FALSE;
+}
+
+
+void
+on_mainwindow_set_focus_child          (GtkContainer    *container,
+                                        GtkWidget       *widget,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(container));
+   if (widget != ibuf->drawing_area)
+      image_buf_focus_lost(ibuf);
+
+}
+
+
+
+void
+on_mainwindow_set_focus                (GtkWindow       *window,
+                                        GtkWidget       *widget,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(window));
+   if (widget != ibuf->drawing_area)
+      image_buf_focus_lost(ibuf);
+
+}
+
+
+void
+on_fontpicker_map                      (GtkWidget       *widget,
+                                        gpointer         user_data)
+{ 
+
+}
+
+
+void
+on_background_color_picker_map         (GtkWidget       *widget,
+                                        gpointer         user_data)
+{
+
+}
+
+
+void
+on_foreground_color_picker_map         (GtkWidget       *widget,
+                                        gpointer         user_data)
+{
+
+}
+
+
+gboolean
+on_fontpicker_map_event                (GtkWidget       *widget,
+                                        GdkEvent        *event,
+                                        gpointer         user_data)
+{
+   return FALSE;
+/*
+   GtkWidget *window = widget_get_toplevel_parent(GTK_WIDGET(widget)); 
+   image_buf *ibuf = widget_get_image(widget);
+   gtk_window_set_transient_for(GTK_WINDOW(window), GTK_WINDOW(ibuf->window));
+   gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
+   gtk_window_set_modal (GTK_WINDOW(window), TRUE);
+  return FALSE;*/
+}
+
+
+gboolean
+on_background_color_picker_map_event   (GtkWidget       *widget,
+                                        GdkEvent        *event,
+                                        gpointer         user_data)
+{
+   return FALSE;
+/*   GtkWidget *window = widget_get_toplevel_parent(GTK_WIDGET(widget)); 
+   image_buf *ibuf = widget_get_image(widget);
+   gtk_window_set_transient_for(GTK_WINDOW(window), GTK_WINDOW(ibuf->window));
+   gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
+   gtk_window_set_modal (GTK_WINDOW(window), TRUE);
+
+  return FALSE;*/
+}
+
+
+gboolean
+on_foreground_color_picker_map_event   (GtkWidget       *widget,
+                                        GdkEvent        *event,
+                                        gpointer         user_data)
+{
+   return FALSE;
+/*   GtkWidget *window = widget_get_toplevel_parent(GTK_WIDGET(widget)); 
+   image_buf *ibuf = widget_get_image(widget);
+   gtk_window_set_transient_for(GTK_WINDOW(window), GTK_WINDOW(ibuf->window));
+   gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
+   gtk_window_set_modal (GTK_WINDOW(window), TRUE);
+
+  return FALSE;*/
+}
+
+
+void
+on_set_as_background_centered_activate (GtkMenuItem     *menuitem,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   image_buf_set_desktop_background(ibuf, 0);
+}
+
+
+void
+on_set_as_background_titled_activate   (GtkMenuItem     *menuitem,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   image_buf_set_desktop_background(ibuf, 1);
+
+}
+
+
+void
+on_foreground_color_picker_color_clicked
+                                        (GtkButton       *button,
+                                        gpointer         user_data)
+{
+
+}
+
+
+void
+on_foreground_color_picker_color_pressed
+                                        (GtkButton       *button,
+                                        gpointer         user_data)
+{
+
+}
+
+
+void
+on_foreground_color_picker_color_released
+                                        (GtkButton       *button,
+                                        gpointer         user_data)
+{
+
+}
+
+
+void
+on_color_palette_entry_map             (GtkWidget       *widget,
+                                        gpointer         user_data)
+{
+
+}
+
+
+void
+on_color_palette_entry_draw            (GtkWidget       *widget,
+                                        GdkRectangle    *area,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(widget);
+   draw_palette_entry(ibuf, widget);
+
+}
+
+
+gboolean
+on_color_palette_entry_key_press_event (GtkWidget       *widget,
+                                        GdkEventKey     *event,
+                                        gpointer         user_data)
+{
+
+  return FALSE;
+}
+
+
+gboolean
+on_color_palette_entry_key_release_event
+                                        (GtkWidget       *widget,
+                                        GdkEventKey     *event,
+                                        gpointer         user_data)
+{
+
+  return FALSE;
+}
+
+
+void
+on_color_palette_entry_realize         (GtkWidget       *widget,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(widget);
+   gchar *name = gtk_widget_get_name(widget);
+}
+
+struct color_selection_info
+{
+   image_buf *ibuf;
+   GtkColorSelection *sel;
+   GtkWidget *dialog;
+};
+
+static void on_color_selector_ok_button_clicked        (GtkButton       *button,
+                                        gpointer         user_data)
+{
+   struct color_selection_info *info = (struct color_selection_info*) (user_data);		
+   GnomeColorPicker *gnomecolorpicker = (GnomeColorPicker*) lookup_widget(info->ibuf->window, "foreground_color_picker");			
+   gdouble tmp[4] = {0};
+   char ctmp[100];
+   GdkColor c = {0};
+   short a;
+   guint32 color;
+   g_assert(info->ibuf);
+   g_assert(gnomecolorpicker);
+   gtk_color_selection_get_color(info->sel, tmp);
+   
+   c.red = (short) (tmp[0] * 65535);
+   c.green = (short) (tmp[1] * 65535);
+   c.blue = (short) (tmp[2] * 65535);
+   a = (short) (tmp[3] * 65535);
+   color = c.red >> 8 << 16 | c.green >> 8 << 8 | c.blue >> 8;
+   gdk_rgb_gc_set_foreground(info->ibuf->gc, color);
+   gnome_color_picker_set_i16(gnomecolorpicker, c.red, c.green, c.blue, a);
+   
+   info->ibuf->palettes[info->ibuf->current_palette] = c;
+   sprintf(ctmp, "color_palette_entry%d", info->ibuf->current_palette);
+   gtk_widget_queue_draw(lookup_widget(info->ibuf->window, ctmp));
+   gtk_widget_destroy(info->dialog);
+   
+   free(info);
+}
+
+static void on_color_selector_cancel_button_clicked        (GtkButton       *button,
+                                        gpointer         user_data)
+{
+   struct color_selection_info *info = (struct color_selection_info*) (user_data);		
+   gtk_widget_destroy(info->dialog);
+}					
+
+gboolean
+on_color_palette_entry_button_press_event
+                                        (GtkWidget       *widget,
+                                        GdkEventButton  *event,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(widget);
+   struct color_selection_info *info;
+   gchar *name = gtk_widget_get_name(widget);
+   int i;
+   sscanf(name, "color_palette_entry%d", &i);
+
+   image_buf_set_foreground_to_palette(ibuf, i);
+   
+   if (event->type == GDK_2BUTTON_PRESS)
+   {
+      GtkWidget *window = gtk_color_selection_dialog_new("Select color");
+      gtk_window_set_transient_for(GTK_WINDOW(window), GTK_WINDOW(ibuf->window));
+      gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
+      gtk_window_set_modal (GTK_WINDOW(window), TRUE);
+      info = calloc(sizeof(struct color_selection_info), 1);
+      info->ibuf = ibuf;
+      info->sel = GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(window)->colorsel);
+      info->dialog = window;
+      gtk_signal_connect(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(window)->ok_button), "clicked", GTK_SIGNAL_FUNC(on_color_selector_ok_button_clicked), (gpointer) info);
+      gtk_signal_connect(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(window)->cancel_button), "clicked", GTK_SIGNAL_FUNC(on_color_selector_cancel_button_clicked), (gpointer) info);
+       
+      gtk_widget_show(window);
+   }
+   
+  return FALSE;
+}
+
+
+gboolean
+on_color_palette_entry_button_release_event
+                                        (GtkWidget       *widget,
+                                        GdkEventButton  *event,
+                                        gpointer         user_data)
+{
+
+  return FALSE;
+}
+
+
+gboolean
+on_color_palette_entry_expose_event    (GtkWidget       *widget,
+                                        GdkEventExpose  *event,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(widget);
+   
+   draw_palette_entry(ibuf, widget);
+  return FALSE;
+}
+
+
+void
+on_select_all_activate                 (GtkMenuItem     *menuitem,
+                                        gpointer         user_data)
+{
+   long w, h;
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   set_drawing_tool(ibuf, LASSO);
+   image_buf_select_all(ibuf);
+   
+}
+   
+
+
+
+void
+on_print_preview_activate              (GtkMenuItem     *menuitem,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   
+   do_print_preview(ibuf);
+   
+
+}
+
+
+void
+on_get_desktop_activate                (GtkMenuItem     *menuitem,
+                                        gpointer         user_data)
+{
+   image_buf *ibuf = widget_get_image(GTK_WIDGET(menuitem));
+   image_buf_get_desktop(ibuf);
+
+}
+
