@@ -2,8 +2,8 @@
 #include <gnome.h>
 #include "pixmaps.h"
 
-#define BEGIN_BUSY_CURSOR { GdkCursor *origcursor = ibuf->cursor; gdk_window_set_cursor(ibuf->window->window, busycursor); gdk_window_set_cursor(ibuf->drawing_area->window, busycursor); gdk_flush(); 
-#define END_BUSY_CURSOR gdk_window_set_cursor(ibuf->window->window, origcursor);  gdk_window_set_cursor(ibuf->drawing_area->window, origcursor); gdk_flush(); }
+#define BEGIN_BUSY_CURSOR { GdkCursor *origcursor = ibuf->cursor; gdk_window_set_cursor(ibuf->window->window, busy_cursor); gdk_window_set_cursor(ibuf->drawing_area->window, busy_cursor); gdk_flush(); 
+#define END_BUSY_CURSOR gdk_window_set_cursor(ibuf->window->window, arrow_cursor);  gdk_window_set_cursor(ibuf->drawing_area->window, origcursor); gdk_flush(); }
 
 void
 on_mainwindow_realize                  (GtkWidget       *widget,
@@ -327,6 +327,16 @@ on_drawingarea_motion_notify_event     (GtkWidget       *widget,
                                         GdkEventMotion  *event,
                                         gpointer         user_data);
 
+gboolean
+on_drawingarea_focus_in_event          (GtkWidget       *widget,
+                                        GdkEventFocus   *event,
+                                        gpointer         user_data);
+gboolean
+on_drawingarea_focus_out_event         (GtkWidget       *widget,
+                                        GdkEventFocus   *event,
+                                        gpointer         user_data);
+
+
 void
 on_drawingarea_realize                 (GtkWidget       *widget,
                                         gpointer         user_data);
@@ -632,3 +642,127 @@ gboolean
 on_drawingarea_key_release_event     (GtkWidget       *widget,
                                         GdkEventKey     *event,
                                         gpointer         user_data);
+
+gboolean
+on_mainwindow_focus_in_event           (GtkWidget       *widget,
+                                        GdkEventFocus   *event,
+                                        gpointer         user_data);
+
+gboolean
+on_mainwindow_focus_out_event          (GtkWidget       *widget,
+                                        GdkEventFocus   *event,
+                                        gpointer         user_data);
+
+void
+on_mainwindow_set_focus_child          (GtkContainer    *container,
+                                        GtkWidget       *widget,
+                                        gpointer         user_data);
+
+
+void
+on_mainwindow_set_focus                (GtkWindow       *window,
+                                        GtkWidget       *widget,
+                                        gpointer         user_data);
+
+void
+on_fontpicker_map                      (GtkWidget       *widget,
+                                        gpointer         user_data);
+
+void
+on_background_color_picker_map         (GtkWidget       *widget,
+                                        gpointer         user_data);
+
+void
+on_foreground_color_picker_map         (GtkWidget       *widget,
+                                        gpointer         user_data);
+
+gboolean
+on_fontpicker_map_event                (GtkWidget       *widget,
+                                        GdkEvent        *event,
+                                        gpointer         user_data);
+
+gboolean
+on_background_color_picker_map_event   (GtkWidget       *widget,
+                                        GdkEvent        *event,
+                                        gpointer         user_data);
+
+gboolean
+on_foreground_color_picker_map_event   (GtkWidget       *widget,
+                                        GdkEvent        *event,
+                                        gpointer         user_data);
+
+void
+on_set_as_background_centered_activate (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_set_as_background_titled_activate   (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_foreground_color_picker_color_clicked
+                                        (GtkButton       *button,
+                                        gpointer         user_data);
+
+void
+on_foreground_color_picker_color_pressed
+                                        (GtkButton       *button,
+                                        gpointer         user_data);
+
+void
+on_foreground_color_picker_color_released
+                                        (GtkButton       *button,
+                                        gpointer         user_data);
+
+void
+on_color_palette_entry_map             (GtkWidget       *widget,
+                                        gpointer         user_data);
+
+void
+on_color_palette_entry_draw            (GtkWidget       *widget,
+                                        GdkRectangle    *area,
+                                        gpointer         user_data);
+
+gboolean
+on_color_palette_entry_key_press_event (GtkWidget       *widget,
+                                        GdkEventKey     *event,
+                                        gpointer         user_data);
+
+gboolean
+on_color_palette_entry_key_release_event
+                                        (GtkWidget       *widget,
+                                        GdkEventKey     *event,
+                                        gpointer         user_data);
+
+void
+on_color_palette_entry_realize         (GtkWidget       *widget,
+                                        gpointer         user_data);
+
+gboolean
+on_color_palette_entry_button_press_event
+                                        (GtkWidget       *widget,
+                                        GdkEventButton  *event,
+                                        gpointer         user_data);
+
+gboolean
+on_color_palette_entry_button_release_event
+                                        (GtkWidget       *widget,
+                                        GdkEventButton  *event,
+                                        gpointer         user_data);
+
+gboolean
+on_color_palette_entry_expose_event    (GtkWidget       *widget,
+                                        GdkEventExpose  *event,
+                                        gpointer         user_data);
+
+void
+on_select_all_activate                 (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_print_preview_activate              (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
+
+void
+on_get_desktop_activate                (GtkMenuItem     *menuitem,
+                                        gpointer         user_data);
