@@ -18,14 +18,20 @@
 */
 
 #include <assert.h>
+#include <sys/param.h>
+#include <gdk/gdkx.h>
 #include "util.h"
 #include "support.h"
 #include "ui.h"
 #include "callbacks.h"
 #include "gtkscrollframe.h"
 #include "pixmaps.h"
+#include "clipboard.h"
 
-GdkCursor *busycursor = 0;
+
+
+GdkCursor *busy_cursor = 0;
+GdkCursor *arrow_cursor = 0;
 
 const int erase_size = 4;
 
@@ -94,8 +100,10 @@ GtkWidget *create_scroll_frame_widget_content(GtkWidget *scrolledwindow1, GtkWid
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (drawingarea);
    gtk_container_add (GTK_CONTAINER (viewport1), drawingarea);
-   gtk_widget_set_events (drawingarea, GDK_EXPOSURE_MASK | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK | GDK_BUTTON_MOTION_MASK | GDK_BUTTON1_MOTION_MASK | GDK_BUTTON2_MOTION_MASK | GDK_BUTTON3_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK | GDK_PROPERTY_CHANGE_MASK | GDK_VISIBILITY_NOTIFY_MASK);
+   gtk_widget_set_events (drawingarea, GDK_EXPOSURE_MASK | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK | GDK_BUTTON_MOTION_MASK | GDK_BUTTON1_MOTION_MASK | GDK_BUTTON2_MOTION_MASK | GDK_BUTTON3_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK | GDK_PROPERTY_CHANGE_MASK | GDK_VISIBILITY_NOTIFY_MASK | GDK_FOCUS_CHANGE_MASK | GDK_STRUCTURE_MASK | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK | GDK_PROPERTY_CHANGE_MASK);
    GTK_WIDGET_SET_FLAGS(drawingarea, GTK_CAN_FOCUS);
+   GTK_WIDGET_SET_FLAGS(drawingarea, GTK_CAN_DEFAULT);
+   
    
    gtk_signal_connect (GTK_OBJECT (drawingarea), "configure_event",
                        GTK_SIGNAL_FUNC (on_drawingarea_configure_event),
@@ -118,6 +126,12 @@ GtkWidget *create_scroll_frame_widget_content(GtkWidget *scrolledwindow1, GtkWid
    gtk_signal_connect (GTK_OBJECT (drawingarea), "delete_event",
                        GTK_SIGNAL_FUNC (on_drawingarea_delete_event),
                        NULL);
+   gtk_signal_connect (GTK_OBJECT (drawingarea), "focus_in_event",
+                       GTK_SIGNAL_FUNC (on_drawingarea_focus_in_event),
+                       NULL);
+   gtk_signal_connect (GTK_OBJECT (drawingarea), "focus_out_event",
+                       GTK_SIGNAL_FUNC (on_drawingarea_focus_out_event),
+                       NULL);
    gtk_signal_connect (GTK_OBJECT (drawingarea), "key_release_event",
                        GTK_SIGNAL_FUNC (on_drawingarea_key_release_event),
                        NULL);
@@ -127,6 +141,18 @@ GtkWidget *create_scroll_frame_widget_content(GtkWidget *scrolledwindow1, GtkWid
 void new_canvas()
 {
    GtkWidget *new_canvas_window = create_new_canvas_window ();
+   GtkWidget *widthentry, *heightentry;
+   char tmp[100];
+   gint position = 0;
+   widthentry = lookup_widget(new_canvas_window, "new_canvas_width_text_entry");
+   heightentry = lookup_widget(new_canvas_window, "new_canvas_height_text_entry");
+   sprintf(tmp, "640");
+   gtk_editable_insert_text(GTK_EDITABLE(widthentry), tmp, strlen(tmp), &position);
+   position = 0; 
+   sprintf(tmp, "480");
+   gtk_editable_insert_text(GTK_EDITABLE(heightentry), tmp, strlen(tmp), &position);
+   
+   
    gtk_widget_show (new_canvas_window);
 }   
 
@@ -135,10 +161,10 @@ static void open_canvas_callback(GtkFileSelection *widget)
    char tmp[4000];
    image_buf *ibuf = (image_buf*) gtk_object_get_user_data(GTK_OBJECT(widget));
    strcpy(tmp, gtk_file_selection_get_filename(GTK_FILE_SELECTION(widget)));
+   gtk_widget_destroy(GTK_WIDGET(widget));
+   PROCESSING_GTK_EVENTS;
    image_buf_load(ibuf, tmp);
    
-   
-   gtk_widget_destroy(GTK_WIDGET(widget));
 }
 
    
@@ -151,15 +177,28 @@ void open_canvas(image_buf *ibuf)
 
    /* And mark it as a transient dialog */
    gtk_window_set_transient_for (GTK_WINDOW(widget), GTK_WINDOW(ibuf->window));
+   gtk_window_set_position(GTK_WINDOW(widget), GTK_WIN_POS_CENTER);
 
    gtk_object_set_user_data(GTK_OBJECT(widget), ibuf);
+   
+   {
+      char pathbuf[MAXPATHLEN], *dir;
+      if (ibuf->name)
+         dir = g_dirname(ibuf->name);
+      else
+         dir = g_get_current_dir();
+
+      sprintf(pathbuf, "%s%c*.*", dir, G_DIR_SEPARATOR);
+      gtk_file_selection_set_filename(GTK_FILE_SELECTION(widget), pathbuf);
+      g_free(dir);
+   }
+   
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(widget)->ok_button),
                              "clicked", GTK_SIGNAL_FUNC(open_canvas_callback),
                              GTK_OBJECT(widget));
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(widget)->cancel_button),
                              "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(widget));
-
    gtk_widget_show(widget);
 
 }
@@ -169,21 +208,31 @@ static void save_canvas_callback(GtkFileSelection *widget)
    image_buf *ibuf = (image_buf*) gtk_object_get_user_data(GTK_OBJECT(widget));
 
    strcpy(tmp, gtk_file_selection_get_filename(GTK_FILE_SELECTION(widget)));
-   image_buf_save(ibuf, tmp);
    gtk_widget_destroy(GTK_WIDGET(widget));
+   PROCESSING_GTK_EVENTS;
+   image_buf_save(ibuf, tmp);
+   image_buf_set_name(ibuf, tmp);
 }
 
-void save_canvas(image_buf *ibuf)
+void save_canvas(image_buf *ibuf, int saveas)
 {
   
-   GtkWidget *widget = gtk_file_selection_new("Save image");
+   GtkWidget *widget ;
+   if ((saveas == 0) && strncmp(ibuf->name, UNTITLED_NAME, strlen(UNTITLED_NAME)))
+   {
+      image_buf_save(ibuf, ibuf->name);
+      return ;
+   }
+   
+   widget = gtk_file_selection_new("Save image");
    g_assert(widget);
    /* Set as modal */
    gtk_window_set_modal (GTK_WINDOW(widget),TRUE);
 
    /* And mark it as a transient dialog */
    gtk_window_set_transient_for(GTK_WINDOW(widget), GTK_WINDOW(ibuf->window));
-   if (ibuf->name)
+   gtk_window_set_position(GTK_WINDOW(widget), GTK_WIN_POS_CENTER);
+   if ((ibuf->name) && strncmp(ibuf->name, UNTITLED_NAME, strlen(UNTITLED_NAME)))
       gtk_file_selection_set_filename(GTK_FILE_SELECTION(widget), ibuf->name);
 
    gtk_object_set_user_data(GTK_OBJECT(widget), ibuf);
@@ -244,15 +293,26 @@ const  toggle_button_info toggle_button_infos[] =
    {"curve_button", CURVE},
    {"oval_button", OVAL},
    {"brush_button", BRUSH},
-   {0, NONE}
+   {0, NONE},
+   {0, PASTE}
 };
+
 void set_drawing_tool(image_buf *ibuf, DRAWING_TOOL tool)
 {
-   const char *buttonname = toggle_button_infos[tool - ERASE].name;
-   GtkWidget *selected = lookup_widget(ibuf->window, buttonname);
+   const char *buttonname = 0;
+   GtkWidget *selected = 0;
    
-   g_assert(selected);
-   select_toolbar_toggle_button(GTK_TOGGLE_BUTTON(selected), 1);
+   if (tool < NONE)
+      buttonname = toggle_button_infos[tool - ERASE].name;
+   
+   if (buttonname)
+      selected = lookup_widget(ibuf->window, buttonname);
+   
+   if (selected)
+      select_toolbar_toggle_button(GTK_TOGGLE_BUTTON(selected), 1);
+   else
+      image_buf_set_tool(ibuf, tool);
+      
 }
 
 
@@ -315,10 +375,88 @@ void StateTimeStep()
 }      
 
 
+static void image_buf_draw_flash(image_buf *ibuf)
+{
+
+   GdkRectangle clientrect ;
+
+   DRAWING_TOOL tool = ibuf->current_tool;
+   GdkGCValues gcvalues;
+   clientrect.x = clientrect.y = 0;
+   clientrect.width = image_buf_width(ibuf);
+   clientrect.height = image_buf_height(ibuf);
+   
+   gdk_gc_get_values(ibuf->gc, &gcvalues);
+   gdk_gc_set_clip_origin(ibuf->gc, 0, 0);
+   gdk_gc_set_clip_rectangle(ibuf->gc, &clientrect);
+   
+   switch(tool)
+   {
+   case TEXT:
+   
+      if ((ibuf->llx != INT_MIN) && (ibuf->lly != INT_MIN))
+      {
+	 gint width, ascent, descent, lbearing, rbearing, junk;
+	 GdkColor white;
+	 gdk_color_white(gdk_rgb_get_cmap(), &white);
+   
+         gdk_gc_set_foreground(ibuf->gc, &white);   
+ 
+	 gdk_string_extents(ibuf->font, "My", &junk, &junk, &junk, &ascent, &descent);
+	 gdk_string_extents(ibuf->font, ibuf->textbuf, &lbearing, &rbearing, &width, &junk, &junk);
+
+	 gdk_gc_set_function(ibuf->gc, GDK_XOR);
+	 gdk_gc_set_line_attributes(ibuf->gc, 2, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);
+	 gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->llx + rbearing + 2, ibuf->lly + descent, ibuf->llx + rbearing + 2, ibuf->lly - ascent);   
+	 gdk_draw_line(ibuf->pixmap, ibuf->gc, ibuf->llx + rbearing + 2, ibuf->lly + descent, ibuf->llx + rbearing + 2, ibuf->lly - ascent);   
+
+	 gdk_gc_set_function(ibuf->gc, gcvalues.function);   
+	 gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, gcvalues.cap_style, gcvalues.join_style);
+	 gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
+	 ibuf->flash_state = !(ibuf->flash_state);
+      }
+      break;
+      
+   case POLSELECT:
+   case LASSO:
+   case PASTE:
+      {
+         gchar dash_length[2] = { 5, 5 };
+	 GdkColor white;
+	 gdk_color_white(gdk_rgb_get_cmap(), &white);
+   
+         gdk_gc_set_foreground(ibuf->gc, &white);   
+ 
+	 gdk_gc_set_function(ibuf->gc, GDK_XOR);
+	 gdk_gc_set_line_attributes(ibuf->gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_ROUND, GDK_JOIN_ROUND);
+         gdk_gc_set_dashes(ibuf->gc,  ibuf->flash_state / 2, dash_length, sizeof(dash_length));
+	 
+	 gdk_draw_lines(ibuf->drawing_area->window, ibuf->gc, ibuf->pts, ibuf->num_pts);
+	 gdk_draw_lines(ibuf->pixmap, ibuf->gc, ibuf->pts, ibuf->num_pts);
+	 
+	 gdk_gc_set_function(ibuf->gc, gcvalues.function);   
+	 gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, gcvalues.cap_style, gcvalues.join_style);
+	 gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
+	 ibuf->flash_state++;
+	 ibuf->flash_state %= 4;
+      }
+      break;
+      
+   default:
+      break;
+   }
+}
+
+void image_buf_clear_flash(image_buf *ibuf)
+{
+   while (ibuf->flash_state)
+      image_buf_draw_flash(ibuf);
+}
+
 void image_buf_set_fill(image_buf *ibuf, int filled)
 {
    GtkWidget *b = lookup_widget(ibuf->window, "filled_button");
-   int old = ibuf->filled;
+   int old = image_buf_get_fill(ibuf);
    assert(b);
    ibuf->filled = filled;
    if (filled != old)
@@ -341,7 +479,7 @@ static void clear_points(image_buf *ibuf)
 
 void image_buf_set_tool(image_buf *ibuf, DRAWING_TOOL tool)
 {
-
+   image_buf_clear_flash(ibuf);
    switch(ibuf->current_tool) /* final mode cleanup */
    {
    case TEXT:
@@ -349,6 +487,7 @@ void image_buf_set_tool(image_buf *ibuf, DRAWING_TOOL tool)
       {
 	 gdk_draw_string(ibuf->drawing_area->window, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
 	 gdk_draw_string(ibuf->pixmap, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
+	 IMAGE_MODIFIED;
       }
       ibuf->llx = INT_MIN;
       ibuf->lly = INT_MIN;
@@ -359,6 +498,7 @@ void image_buf_set_tool(image_buf *ibuf, DRAWING_TOOL tool)
    }
    
    ibuf->current_tool = tool;
+   ibuf->flash_state = 0;
    switch(tool) /* new mode initialization */
    {
    case MULTILINE:
@@ -368,6 +508,8 @@ void image_buf_set_tool(image_buf *ibuf, DRAWING_TOOL tool)
    case ARC:
    case RECTANGLE:
    case OVAL:
+   case LASSO:
+   case POLSELECT:
    
       image_buf_set_cursor(ibuf, gdk_cursor_new(GDK_CROSSHAIR));
       break;
@@ -392,18 +534,33 @@ void image_buf_set_tool(image_buf *ibuf, DRAWING_TOOL tool)
    ibuf->ly = INT_MIN;
    ibuf->mx = INT_MIN;
    ibuf->my = INT_MIN;
+   ibuf->lpx = INT_MIN;
+   ibuf->lpy = INT_MIN;
    clear_points(ibuf);
 }
    
 
 static void insert_point(image_buf *ibuf, int x, int y)
 {
+   if (x < 0)
+      x = 0;
+   if (y < 0)
+      y = 0;
    ibuf->pts[ibuf->num_pts].x = x;
    ibuf->pts[ibuf->num_pts].y = y;
    ibuf->num_pts++;
 }
 
-static void draw_rectangle(GdkDrawable *d, GdkGC *gc, gint filled, gint x1, gint y1, gint x2, gint y2)
+
+static void modify_point(image_buf *ibuf, int index, int x, int y)
+{
+   ibuf->pts[index].x = x;
+   ibuf->pts[index].y = y;
+
+
+}
+
+void draw_rectangle(GdkDrawable *d, GdkGC *gc, gint filled, gint x1, gint y1, gint x2, gint y2)
 {
    int ox, w, oy, h;
    if (x1 > x2)
@@ -517,6 +674,35 @@ static void draw_arc(GdkDrawable *d, GdkGC *gc, gint filled, gint x1, gint y1, g
    gdk_draw_arc(d, gc, filled, ox, oy, w, h, angle1 * 64, angle2 * 64);
 }
 
+
+
+void image_buf_clear(image_buf *ibuf)
+{
+   GdkGCValues gcvalues;
+   GdkRectangle clientrect ;
+      
+   
+   clientrect.x = clientrect.y = 0;
+   clientrect.width = image_buf_width(ibuf);
+   clientrect.height = image_buf_height(ibuf);
+   
+   image_buf_clear_flash(ibuf);
+   
+   gdk_gc_get_values(ibuf->gc, &gcvalues);
+   gdk_gc_set_clip_origin(ibuf->gc, 0, 0);
+   gdk_gc_set_clip_rectangle(ibuf->gc, &clientrect);
+   
+   gdk_gc_set_foreground(ibuf->gc, &(gcvalues.background));
+
+   draw_rectangle(ibuf->pixmap, ibuf->gc, TRUE, clientrect.x, clientrect.y, clientrect.width, clientrect.height);
+   draw_rectangle(ibuf->drawing_area->window, ibuf->gc, TRUE, clientrect.x, clientrect.y, clientrect.width, clientrect.height);
+   IMAGE_MODIFIED;
+   
+   gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
+
+}
+
+
 void handle_button_press(image_buf *ibuf, int x, int y)
 {
    DRAWING_TOOL tool = ibuf->current_tool;
@@ -535,7 +721,25 @@ void handle_button_press(image_buf *ibuf, int x, int y)
    gdk_gc_set_clip_origin(ibuf->gc, 0, 0);
    gdk_gc_set_clip_rectangle(ibuf->gc, &clientrect);
    
-   insert_point(ibuf, x, y);
+   image_buf_clear_flash(ibuf);
+   
+   if (tool == PASTE)
+      ;
+   
+   else if (tool != POLSELECT)
+   {
+      clear_points(ibuf);
+      insert_point(ibuf, x, y);
+   }
+   else
+   {
+      if (ibuf->num_pts)
+         modify_point(ibuf, ibuf->num_pts - 1, x, y);
+      else
+         insert_point(ibuf, x, y);
+      
+   }
+   
    switch (tool)
    {
    case CLOSED_FREEHAND:
@@ -543,15 +747,21 @@ void handle_button_press(image_buf *ibuf, int x, int y)
       gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, GDK_CAP_ROUND, GDK_JOIN_ROUND);
       gdk_draw_line(ibuf->pixmap, ibuf->gc, x, y, x, y);
       gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, x, y, x, y);
+      IMAGE_MODIFIED;
+
       gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, gcvalues.cap_style, gcvalues.join_style);
       //gtk_widget_draw(ibuf->drawing_area, &rect);
       break;
+    
+     
    
    case MULTILINE:
       if ((ibuf->llx > INT_MIN) && (ibuf->lly > INT_MIN))
       {
 	 gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, x, y, ibuf->llx, ibuf->lly);
 	 gdk_draw_line(ibuf->pixmap, ibuf->gc, x, y, ibuf->llx, ibuf->lly);
+	 IMAGE_MODIFIED;
+
       }      
       
    case LINE:
@@ -566,13 +776,30 @@ void handle_button_press(image_buf *ibuf, int x, int y)
    case RECTANGLE:
       break;	 
 	
+   case LASSO:
+   case POLSELECT:
+      break;	
+	
    case TEXT:
+      if ((ibuf->llx != INT_MIN) && (ibuf->lly != INT_MIN)) /* previous round of text entry */
+      {
+	 if (strlen(ibuf->textbuf))
+	 {
+	    gdk_draw_string(ibuf->drawing_area->window, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
+	    gdk_draw_string(ibuf->pixmap, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
+   	    IMAGE_MODIFIED;
+	 }
+	 strcpy(ibuf->textbuf, "");
+	 ibuf->llx = INT_MIN;
+	 ibuf->lly = INT_MIN;
+      }
       break;
    	
    case ERASE:
       gdk_gc_set_foreground(ibuf->gc, &(gcvalues.background));
       draw_rectangle(ibuf->pixmap, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
       draw_rectangle(ibuf->drawing_area->window, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
+      IMAGE_MODIFIED;
       
       gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
       break;
@@ -582,6 +809,11 @@ void handle_button_press(image_buf *ibuf, int x, int y)
       gtk_widget_draw(ibuf->drawing_area, NULL);
       break;
       	 
+       
+   case PASTE:
+      image_buf_put_clipboard_image(ibuf, x, y);
+      break;
+                
    default:
       gdk_beep();
    }
@@ -612,8 +844,22 @@ void handle_button_move(image_buf *ibuf, int x, int y)
    gdk_gc_set_clip_origin(ibuf->gc, 0, 0);
    gdk_gc_set_clip_rectangle(ibuf->gc, &clientrect);
    
-   insert_point(ibuf, x, y);
-   
+   image_buf_clear_flash(ibuf);
+   if (tool == PASTE)
+      ;
+   else
+   if (tool != POLSELECT)
+      insert_point(ibuf, x, y);
+   else
+   {
+      if ((ibuf->mx == INT_MIN) || (ibuf->my == INT_MIN))
+	 insert_point(ibuf, x, y);
+      else
+         modify_point(ibuf, ibuf->num_pts - 1, x, y); 
+      
+   }
+   image_buf_draw_flash(ibuf);
+         
    switch (tool)
    {
    case CLOSED_FREEHAND:
@@ -630,9 +876,12 @@ void handle_button_move(image_buf *ibuf, int x, int y)
 	 gdk_draw_line(ibuf->pixmap, ibuf->gc, ibuf->mx, ibuf->my, x, y);
 	 gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->mx, ibuf->my, x, y);
       }
+      IMAGE_MODIFIED;
       gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, gcvalues.cap_style, gcvalues.join_style);
       //gtk_widget_draw(ibuf->drawing_area, &rect);
       break;
+   
+   
       
    case MULTILINE:
    case LINE:
@@ -643,6 +892,7 @@ void handle_button_move(image_buf *ibuf, int x, int y)
       }
       gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
       gdk_gc_set_function(ibuf->gc, gcvalues.function);  
+      gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, gcvalues.cap_style, gcvalues.join_style);
       break;
       
    case OVAL:
@@ -680,10 +930,16 @@ void handle_button_move(image_buf *ibuf, int x, int y)
       gdk_gc_set_foreground(ibuf->gc, &(gcvalues.background));
       draw_rectangle(ibuf->pixmap, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
       draw_rectangle(ibuf->drawing_area->window, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
-      
+      IMAGE_MODIFIED;
+
       gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
       break;
        
+       
+   case PASTE:
+      image_buf_put_clipboard_image(ibuf, x, y);
+      break;
+                
    default:
       
    }
@@ -711,7 +967,23 @@ void handle_button_release(image_buf *ibuf, int x, int y)
    gdk_gc_get_values(ibuf->gc, &gcvalues);
    gdk_gc_set_clip_origin(ibuf->gc, 0, 0);
    gdk_gc_set_clip_rectangle(ibuf->gc, &clientrect);
-   insert_point(ibuf, x, y);
+   
+   image_buf_clear_flash(ibuf);
+   
+   if (tool == PASTE)
+      ;
+   else
+   if (tool != POLSELECT)
+      insert_point(ibuf, x, y);
+   else
+   {
+      if ((ibuf->mx == INT_MIN) || (ibuf->my == INT_MIN))
+	 insert_point(ibuf, x, y);
+      else
+         modify_point(ibuf, ibuf->num_pts - 1, x, y); 
+      
+   }
+    
    
    switch (tool)
    {
@@ -722,8 +994,8 @@ void handle_button_release(image_buf *ibuf, int x, int y)
       {
          if (tool == CLOSED_FREEHAND)
 	 {
-	    gdk_draw_polygon(ibuf->pixmap, ibuf->gc, ibuf->filled, ibuf->pts, ibuf->num_pts);
-	    gdk_draw_polygon(ibuf->drawing_area->window, ibuf->gc, ibuf->filled, ibuf->pts, ibuf->num_pts);
+	    gdk_draw_polygon(ibuf->pixmap, ibuf->gc, image_buf_get_fill(ibuf), ibuf->pts, ibuf->num_pts);
+	    gdk_draw_polygon(ibuf->drawing_area->window, ibuf->gc, image_buf_get_fill(ibuf), ibuf->pts, ibuf->num_pts);
 	 
 	 }
 	 else
@@ -731,19 +1003,22 @@ void handle_button_release(image_buf *ibuf, int x, int y)
 	    gdk_draw_line(ibuf->pixmap, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
 	    gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
 	 }
+      	 IMAGE_MODIFIED;
       }
       else
       {
          if (tool == CLOSED_FREEHAND)
 	 {
-	    gdk_draw_polygon(ibuf->pixmap, ibuf->gc, ibuf->filled, ibuf->pts, ibuf->num_pts);
-	    gdk_draw_polygon(ibuf->drawing_area->window, ibuf->gc, ibuf->filled, ibuf->pts, ibuf->num_pts);
+	    gdk_draw_polygon(ibuf->pixmap, ibuf->gc, image_buf_get_fill(ibuf), ibuf->pts, ibuf->num_pts);
+	    gdk_draw_polygon(ibuf->drawing_area->window, ibuf->gc, image_buf_get_fill(ibuf), ibuf->pts, ibuf->num_pts);
 	 }
 	 else
 	 {
 	    gdk_draw_line(ibuf->pixmap, ibuf->gc, ibuf->mx, ibuf->my, x, y);
 	    gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->mx, ibuf->my, x, y);
 	 }
+	 IMAGE_MODIFIED;
+
       }
       gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, gcvalues.cap_style, gcvalues.join_style);
       //gtk_widget_draw(ibuf->drawing_area, &rect);
@@ -758,6 +1033,7 @@ void handle_button_release(image_buf *ibuf, int x, int y)
       gdk_gc_set_function(ibuf->gc, gcvalues.function);   
       gdk_draw_line(ibuf->pixmap, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
       gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
+      IMAGE_MODIFIED;
       break;
       
    case OVAL:
@@ -766,8 +1042,9 @@ void handle_button_release(image_buf *ibuf, int x, int y)
          draw_oval(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
       
       gdk_gc_set_function(ibuf->gc, gcvalues.function);   
-      draw_oval(ibuf->pixmap, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
-      draw_oval(ibuf->drawing_area->window, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
+      draw_oval(ibuf->pixmap, ibuf->gc, image_buf_get_fill(ibuf), ibuf->lx, ibuf->ly, x, y);
+      draw_oval(ibuf->drawing_area->window, ibuf->gc, image_buf_get_fill(ibuf), ibuf->lx, ibuf->ly, x, y);
+      IMAGE_MODIFIED;
       break;
    
    case ARC:
@@ -776,8 +1053,9 @@ void handle_button_release(image_buf *ibuf, int x, int y)
          draw_arc(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
       
       gdk_gc_set_function(ibuf->gc, gcvalues.function);   
-      draw_arc(ibuf->pixmap, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
-      draw_arc(ibuf->drawing_area->window, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
+      draw_arc(ibuf->pixmap, ibuf->gc, image_buf_get_fill(ibuf), ibuf->lx, ibuf->ly, x, y);
+      draw_arc(ibuf->drawing_area->window, ibuf->gc, image_buf_get_fill(ibuf), ibuf->lx, ibuf->ly, x, y);
+      IMAGE_MODIFIED;
       break;
       
       
@@ -787,8 +1065,9 @@ void handle_button_release(image_buf *ibuf, int x, int y)
          draw_rectangle(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
       
       gdk_gc_set_function(ibuf->gc, gcvalues.function);   
-      draw_rectangle(ibuf->pixmap, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
-      draw_rectangle(ibuf->drawing_area->window, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
+      draw_rectangle(ibuf->pixmap, ibuf->gc, image_buf_get_fill(ibuf), ibuf->lx, ibuf->ly, x, y);
+      draw_rectangle(ibuf->drawing_area->window, ibuf->gc, image_buf_get_fill(ibuf), ibuf->lx, ibuf->ly, x, y);
+      IMAGE_MODIFIED;
       break;	 
    
   
@@ -796,14 +1075,22 @@ void handle_button_release(image_buf *ibuf, int x, int y)
       gdk_gc_set_foreground(ibuf->gc, &(gcvalues.background));
       draw_rectangle(ibuf->pixmap, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
       draw_rectangle(ibuf->drawing_area->window, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
-      
+      IMAGE_MODIFIED;
+
       gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
       break;
          
    case TEXT:
       gtk_widget_grab_focus(ibuf->drawing_area);
+      gtk_widget_grab_default(ibuf->drawing_area);
       break;
-      
+   
+   
+   
+   case PASTE:
+      image_buf_put_clipboard_image(ibuf, x, y);
+      break;
+         
    default:
       
    }
@@ -817,22 +1104,46 @@ void handle_button_release(image_buf *ibuf, int x, int y)
       ibuf->lly = y;
    }
    
+   if (tool == PASTE)
+      ;
+   else
    if (tool == TEXT)
    {
       ibuf->llx = x;
       ibuf->lly = y;
    }
-      
-   clear_points(ibuf);
+   else if (tool == LASSO)
+   {
+      insert_point(ibuf, ibuf->pts[0].x, ibuf->pts[0].y);
+   }
+   else if (tool == POLSELECT)
+   {
+      insert_point(ibuf, ibuf->pts[0].x, ibuf->pts[0].y);
+   }
+   else
+      clear_points(ibuf);
 }   
 
 
 void handle_key_release(image_buf *ibuf, GdkEventKey *keyevent)
 {
    GdkGCValues gcvalues;
-   
+   GdkColor white;
    int length;
    char tmp[MAX_TEXT_LENGTH];
+   
+   if ((ibuf->current_tool == PASTE))
+   {
+      if (keyevent->keyval == GDK_Escape)
+      {
+         image_buf_clear_flash(ibuf);
+	 image_buf_erase_dynamic_image(ibuf, clipboard, ibuf->lpx, ibuf->lpy);
+	 image_buf_leave_paste_mode(ibuf);
+	 return;
+      }
+   }
+      
+   
    if ((ibuf->llx == INT_MIN) || (ibuf->lly == INT_MIN) || (ibuf->current_tool != TEXT))
    {
       gdk_beep();
@@ -840,6 +1151,9 @@ void handle_key_release(image_buf *ibuf, GdkEventKey *keyevent)
    }
    
    
+   
+   image_buf_clear_flash(ibuf);
+   
    gdk_gc_get_values(ibuf->gc, &gcvalues);
    strcpy(tmp, ibuf->textbuf);
    length = strlen(ibuf->textbuf);
@@ -850,13 +1164,19 @@ void handle_key_release(image_buf *ibuf, GdkEventKey *keyevent)
    }
    else if (keyevent->string)
       strcat(ibuf->textbuf, keyevent->string);
-   else if (keyevent->keyval >= GDK_space)
+   else if ((keyevent->keyval >= GDK_space) && (keyevent->keyval < GDK_Shift_L))
    {
       sprintf(ibuf->textbuf, "%s%c", tmp, keyevent->keyval);
    }
+   if (keyevent->keyval == GDK_Escape)
+   {
+      strcpy(ibuf->textbuf, "");
+   }
    
    gdk_gc_set_function(ibuf->gc, GDK_XOR);
-      
+   gdk_color_white(gdk_rgb_get_cmap(), &white);
+   
+   gdk_gc_set_foreground(ibuf->gc, &white);   
    if (strlen(tmp))
    {
       gdk_draw_string(ibuf->drawing_area->window, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, tmp);
@@ -871,19 +1191,212 @@ void handle_key_release(image_buf *ibuf, GdkEventKey *keyevent)
    
    
    gdk_gc_set_function(ibuf->gc, gcvalues.function);   
-
+   gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
+   
    if (keyevent->keyval == GDK_Return)
    {
       if (strlen(ibuf->textbuf))
       {
 	 gdk_draw_string(ibuf->drawing_area->window, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
 	 gdk_draw_string(ibuf->pixmap, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
+	 IMAGE_MODIFIED;
+
       }
       strcpy(ibuf->textbuf, "");
-      ibuf->llx = INT_MIN;
-      ibuf->lly = INT_MIN;
+      
+      ibuf->lly +=  (int) (gdk_string_height(ibuf->font, "My") * 1.1 + 0.5);
+   }
+   else if (keyevent->keyval == GDK_Escape)
+   {
+      ibuf->llx = ibuf->lly = INT_MIN;
+   }
+}
+
+
+int handle_timeout(image_buf *ibuf)
+{
+   if (image_buf_has_focus(ibuf)) 
+      image_buf_draw_flash(ibuf);
+   else
+      image_buf_clear_flash(ibuf);
+   return 1;
+}
+
+void image_buf_select_all(image_buf *ibuf)
+{
+   long w, h;
+   
+   w = image_buf_width(ibuf);
+   h = image_buf_height(ibuf);
+   
+   image_buf_clear_flash(ibuf);
+   
+   clear_points(ibuf);
+   insert_point(ibuf, 0, 0);
+   insert_point(ibuf, w - 1, 0);
+   insert_point(ibuf, w - 1, h - 1);
+   insert_point(ibuf, 0, h - 1);
+}
+
+
+void image_buf_save_tool(image_buf *ibuf)
+{
+   ibuf->saved_tool = ibuf->current_tool;
+}
+
+void image_buf_restore_tool(image_buf *ibuf)
+{
+   set_drawing_tool(ibuf, ibuf->saved_tool);
+}
+
+void image_buf_enter_paste_mode(image_buf *ibuf)
+{
+   image_buf_save_tool(ibuf);
+   set_drawing_tool(ibuf, PASTE);
+}
+
+void image_buf_leave_paste_mode(image_buf *ibuf)
+{
+   image_buf_pixmap_to_rgbbuf(ibuf, 0);
+   IMAGE_MODIFIED;
+   image_buf_restore_tool(ibuf);
+}
+
+
+
+void image_buf_draw_dynamic_image(image_buf *ibuf, GdkPixbuf *new_image, int x, int y)
+{
+   GdkRectangle rect ;
+   
+   if (!new_image)
+   {
+      gdk_beep();
+      return;
    }
    
+   rect.x = x;
+   rect.y = y;
+   rect.width = gdk_pixbuf_get_width(new_image);
+   rect.height = gdk_pixbuf_get_height(new_image);
+   
+   if (x + rect.width >= image_buf_width(ibuf))
+      rect.width = image_buf_width(ibuf) - x - 1;
+   
+   if (y + rect.height >= image_buf_height(ibuf))
+      rect.height = image_buf_height(ibuf) - y - 1;
+   
+   if ((image_buf_width(ibuf) - x - 1 < 0) || (image_buf_height(ibuf) - y - 1 < 0))
+      return;
+      
+   image_buf_clear_flash(ibuf);
+      
+   
+   image_buf_rgbbuf_to_pixmap(ibuf, &rect);
+   
+   gdk_pixbuf_render_to_drawable_alpha(new_image, ibuf->pixmap, 0, 0, x, y, rect.width, rect.height, GDK_PIXBUF_ALPHA_BILEVEL, 255 / 2, GDK_RGB_DITHER_MAX, 0, 0);
+   gtk_widget_queue_draw_area(ibuf->drawing_area, rect.x, rect.y, rect.width, rect.height);
+   gtk_widget_draw(ibuf->drawing_area, &rect);
+}
+
+
+
+void image_buf_erase_dynamic_image(image_buf *ibuf, GdkPixbuf *new_image, int x, int y)
+{
+   GdkRectangle rect ;
+   
+   if (!new_image)
+   {
+      gdk_beep();
+      return;
+   }
+   
+   rect.x = x;
+   rect.y = y;
+   rect.width = gdk_pixbuf_get_width(new_image);
+   rect.height = gdk_pixbuf_get_height(new_image);
+   
+   if (x + rect.width >= image_buf_width(ibuf))
+      rect.width = image_buf_width(ibuf) - x - 1;
+   
+   if (y + rect.height >= image_buf_height(ibuf))
+      rect.height = image_buf_height(ibuf) - y - 1;
+   
+   if ((image_buf_width(ibuf) - x - 1 < 0) || (image_buf_height(ibuf) - y - 1 < 0))
+      return;
+      
+   image_buf_clear_flash(ibuf);
+   
+   image_buf_rgbbuf_to_pixmap(ibuf, &rect);
+   gtk_widget_queue_draw_area(ibuf->drawing_area, rect.x, rect.y, rect.width, rect.height);
+}   
+
+void image_buf_put_clipboard_image(image_buf *ibuf, int x, int y)
+{
+   if ((ibuf->lpx != INT_MIN) && (ibuf->lpy != INT_MIN))
+   {
+      image_buf_erase_dynamic_image(ibuf, clipboard, ibuf->lpx, ibuf->lpy);
+   }
+   image_buf_draw_dynamic_image(ibuf, clipboard, x, y);
+   image_buf_set_pts_top_left(ibuf, x, y); 
+   ibuf->lpx = x;
+   ibuf->lpy = y;   
+   /*gtk_widget_grab_focus(ibuf->drawing_area);*/
+
+}   
+
+
+void image_buf_get_desktop(image_buf *ibuf)
+{
+
+   GdkWindow *root = (GdkWindow *) &gdk_root_parent;
+   int w, h;
+   
+   gdk_window_get_size (root, &w, &h);
+   image_buf_resize(ibuf, w, h);
+   gdk_pixbuf_get_from_drawable(ibuf->rgbbuf, root, 0, 0, 0, 0, 0, w, h);
+   image_buf_rgbbuf_to_pixmap(ibuf, 0);
+   gtk_widget_draw(ibuf->drawing_area, NULL);
+}
+   
+
+
+void image_buf_set_desktop_background(image_buf* ibuf, int type)
+{/* type: 0:centered, 1: tiled */ /* not working yet */
+
+   GdkWindow *root = (GdkWindow *) &gdk_root_parent;
+   int w, h;
+   
+   gdk_window_get_size (root, &w, &h);
+   gdk_window_set_back_pixmap(root, ibuf->pixmap, FALSE);
+   gdk_window_clear(root);
+   
+}
+
+
+void draw_palette_entry(image_buf *ibuf, GtkWidget *palette_entry)
+{
+   GdkRectangle clientrect ;
+   gchar *name = gtk_widget_get_name(palette_entry);
+   guint32 color;
+   int i, n;
+   GdkGCValues gcvalues;
+   
+   i = sscanf(name, "color_palette_entry%d", &n);
+   assert(i);
+   
+   clientrect.x = clientrect.y = 0;
+   clientrect.width = palette_entry->allocation.width;
+   clientrect.height = palette_entry->allocation.height;
+   
+   gdk_gc_get_values(ibuf->gc, &gcvalues);
+   gdk_gc_set_clip_origin(ibuf->gc, 0, 0);
+   gdk_gc_set_clip_rectangle(ibuf->gc, &clientrect);
+   color = ibuf->palettes[n].red >> 8 << 16 | ibuf->palettes[n].green >> 8 << 8 | ibuf->palettes[n].blue >> 8;
+   
+   gdk_rgb_gc_set_foreground(ibuf->gc, color);
+   draw_rectangle(palette_entry->window, ibuf->gc, TRUE,   clientrect.x, clientrect.y, clientrect.x + clientrect.width, clientrect.y + clientrect.height);
+   gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
+  
 }
 
 
@@ -905,9 +1418,9 @@ void GdkColor_to_rgb(const GdkColor *color, unsigned char *r, unsigned char *g,
       break;
    case GDK_VISUAL_TRUE_COLOR:
    case GDK_VISUAL_DIRECT_COLOR:
-      *r = (color->pixel & visual->red_mask) >> visual->red_shift;
-      *g = (color->pixel & visual->green_mask) >> visual->green_shift;
-      *b = (color->pixel & visual->blue_mask) >> visual->blue_shift;
+      *r = (color->pixel & visual->red_mask) >> visual->red_shift << (sizeof(unsigned char) * 8 - visual->red_prec);
+      *g = (color->pixel & visual->green_mask) >> visual->green_shift << (sizeof(unsigned char) * 8 - visual->green_prec);
+      *b = (color->pixel & visual->blue_mask) >> visual->blue_shift << (sizeof(unsigned char) * 8 - visual->blue_prec);
       break;
    
    default:
@@ -922,4 +1435,34 @@ void spreadimage(image_buf * input, image_buf *output)
 {
    gdk_pixbuf_copy_area(input->rgbbuf, 0, 0, image_buf_width(input), image_buf_height(input), output->rgbbuf, 0, 0);
    ImageSpread(output);
+   image_buf_set_modified(output, 1);
+ 
 }
+
+
+
+GdkRectangle compute_cover_rect(GdkPoint *pts, int num_pts)
+{
+   
+   GdkRectangle rect;
+   int i;
+   int x = INT_MAX, y = INT_MAX, x2 = INT_MIN, y2 = INT_MIN;
+   assert(num_pts);
+   for (i = 0; i < num_pts; i++)
+   {
+      if (x > pts[i].x)
+         x = pts[i].x;
+      if (y > pts[i].y)
+         y = pts[i].y;
+      if (x2 < pts[i].x)
+         x2 = pts[i].x;
+      if (y2 < pts[i].y)
+         y2 = pts[i].y;
+   }
+   rect.x = x;
+   rect.y = y;
+   rect.width = x2 - x + 1;
+   rect.height = y2 - y + 1;
+   return rect;
+}
+   
